"""Shows the source visibility at a given location and time."""

import datetime
from collections import namedtuple
from typing import Literal

import astropy.units as u
import dateutil.parser
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from astropy.coordinates import AltAz, BaseCoordinateFrame, EarthLocation, SkyCoord
from astropy.time import Time
from rich.console import Console
from rich.table import Table

from radiotools.layouts import Layout
from radiotools.utils import get_array_names

COLORS = [
    "#fd7f6f",
    "#7eb0d5",
    "#b2e061",
    "#bd7ebe",
    "#ffb55a",
    "#ffee65",
    "#beb9db",
    "#fdcce5",
    "#8bd3c7",
    "#b3d4ff",
]

PYVISGEN_LAYOUTS = "https://raw.githubusercontent.com/radionets-project/"
PYVISGEN_LAYOUTS += "pyvisgen/refs/heads/main/resources/layouts/"
PYVISGEN = "https://github.com/radionets-project/pyvisgen/tree/main/resources/layouts"

LOCATIONS = [
    "alma",
    "alma_dsharp",
    "dsa2000W",
    "dsa2000_31b",
    "eht",
    "meerkat",
    "vla",
    "vlba",
]


class SourceVisibility:
    """Plots the source visibility for a given location
    and time range.

    Parameters
    ----------
    target : tuple or str
        Tuple of RA/Dec coordinates or string of valid
        target name (ICRS).
    date : str or list[str]
        Date or start and end points of a range of dates.
    location : str or :class:`astropy.coordinates.EarthLocation`, optional
        Name of an existing array layout included in pyvisgen,
        a location, or :class:`astropy.coordinates.EarthLocation`
        object of an observatory or telescope. Default: ``None``
    obs_length: float, optional
        Observation length in hours. Default: ``None``
    frame : str or :class:`astropy.coordinates.BaseCoordinateFrame`, optional
        Type of coordinate frame the source sky coordinates
        should represent. Defaults: ``'icrs'``
    print_optimal_date : bool, optional
        Prints the optimal date to observe. Default: False
    """

    def __init__(
        self,
        target: tuple | str,
        date: str | list[str],
        location: str | EarthLocation = None,
        obs_length: float = 4.0,
        frame: str | BaseCoordinateFrame = "icrs",
        min_alt: float = 15.0,
        max_alt: float = 85.0,
        print_optimal_date: bool = False,
    ) -> None:
        """Initializes the class with source and observation information.

        Parameters
        ----------
        target : tuple or str
            Tuple of RA/Dec coordinates or string of valid
            target name (ICRS).
        date : str or list[str]
            Date or start and end points of a range of dates.
        location : str or :class:`astropy.coordinates.EarthLocation`, optional
            Name of an existing array layout included in pyvisgen,
            a location, or :class:`astropy.coordinates.EarthLocation` object of an
            observatory or telescope. Default: ``None``
        obs_length: float, optional
            Observation length in hours. Default: ``None``
        frame : str or :class:`astropy.coordinates.BaseCoordinateFrame`, optional
            Type of coordinate frame the source sky coordinates
            should represent. Defaults: 'icrs'
        min_alt : float, optional
            The minimum altitude of the source (in degrees) above the horizon to
            be determined as visible. Default: 15.0
        max_alt : float, optional
            The maximum altitude of the source (in degrees) above the horizon to
            be determined as visible. Default: 85.0
        print_optimal_date: bool, optional
            If `True` prints the optimal date for the observation. Default: ``False``
        """
        if isinstance(target, tuple) or (
            isinstance(target, np.ndarray) and target.size == 2
        ):
            self.ra = u.Quantity(target[0], unit=u.deg)
            self.dec = u.Quantity(target[1], unit=u.deg)
            self.source = SkyCoord(self.ra, self.dec, frame=frame)
            self.target_name = None

        elif isinstance(target, str):
            self.source = SkyCoord.from_name(target, frame=frame)
            self.ra = self.source.ra
            self.dec = self.source.dec
            self.target_name = target

        else:
            raise TypeError(
                "Please either provide a valid target name or a RA/Dec tuple!"
            )

        if isinstance(location, str) and (
            (location in get_array_names(PYVISGEN)) or (location in LOCATIONS)
        ):
            self.name = location
            self.array = Layout.from_url(PYVISGEN_LAYOUTS + location + ".txt")

            self.location = EarthLocation.from_geocentric(
                self.array.x, self.array.y, self.array.z, unit=u.m
            )

        elif isinstance(location, str):
            self.name = location
            self.location = EarthLocation.of_address(location)

        elif isinstance(location, EarthLocation):
            self.name = location
            self.location = location

        elif isinstance(location, tuple) and list(map(type, location)) == [
            str,
            pd.DataFrame,
        ]:
            self.name = location[0]
            self.location = EarthLocation.from_geocentric(
                location[1].X,
                location[1].Y,
                location[1].Z,
                unit=u.m,
            )

        else:
            raise ValueError("Please provide a valid location!")

        self.date = date
        self.obs_length = obs_length

        if min_alt > max_alt:
            raise ValueError(
                "Expected 'min_alt' to be smaller than 'max_alt', but got "
                f"min_alt = {min_alt} and max_alt = {max_alt}."
            )

        if min_alt < 0 or min_alt > 90 or max_alt < 0 or max_alt > 90:
            raise ValueError(
                "Expected 'min_alt' and 'max_alt' to be between 0 and 90 degrees."
            )

        self.min_alt = min_alt
        self.max_alt = max_alt

        self._get_dates()
        self._get_pos()

        self.get_optimal_date(print_optimal_date)

    def _get_dates(self) -> None:
        """Creates a date range from the dates given
        when initializing this class.
        """
        if isinstance(self.date, str):
            start_date = dateutil.parser.parse(self.date)
            end_date = start_date + datetime.timedelta(1)
        else:
            start_date = dateutil.parser.parse(self.date[0])
            end_date = dateutil.parser.parse(self.date[1])

        self.dates = pd.date_range(start_date, end_date, periods=1000).strftime(
            "%Y-%m-%d %H:%M:%S"
        )
        self.dates = pd.to_datetime(self.dates)

    def _get_pos(self) -> None:
        """Creates the sky coordinates of the source
        throughout the time range for a given frame
        (default is ICRS) and calculates the altitude
        and azimuth in the AltAz frame.
        """
        self.source_pos = dict()

        if self.location.size == 1:
            altaz_frame = AltAz(obstime=Time(self.dates), location=self.location)
            self.source_pos[0] = self.source.transform_to(altaz_frame)

        elif self.location.size > 10:
            altaz_frame = AltAz(obstime=Time(self.dates), location=self.location[0])
            self.source_pos[0] = self.source.transform_to(altaz_frame)

        else:
            for i, loc in enumerate(self.location):
                altaz_frame = AltAz(obstime=Time(self.dates), location=loc)
                self.source_pos[i] = self.source.transform_to(altaz_frame)

    def _plot_config(self, ax) -> None:
        """Settings for the plot.

        Parameters
        ----------
        ax : matplotlib.axes.Axes.axis
            Axis object of the current figure.
        """
        locale_label_strings = {
            "EN": dict(xlabel="Time / hours", ylabel="Altitude / deg"),
            "DE": dict(xlabel="Zeit / Stunden", ylabel="Höhenwinkel / Grad"),
        }

        locale_legend_strings = {
            "EN": dict(title="Telescope ID"),
            "DE": dict(title="Teleskop-Nummer"),
        }

        # axis ticks
        ax["A"].set_xticks(
            ax["A"].get_xticks(), ax["A"].get_xticklabels(), rotation=45, ha="right"
        )

        # further axis settings
        ax["A"].set(
            **locale_label_strings[self.plot_locale],
            ylim=(0, 90),
            title=self.name,
        )
        ax["A"].xaxis.set_major_formatter(
            mdates.ConciseDateFormatter(ax["A"].xaxis.get_major_locator())
        )
        ax["B"].axis("off")

        # legend is plotted on axis "B"
        if self.legend:
            handles, labels = ax["A"].get_legend_handles_labels()
            ax["B"].legend(
                ncols=2,
                handles=handles,
                labels=labels,
                loc="upper left",
                handlelength=4,
                **locale_legend_strings[self.plot_locale],
            )

        # Text is drawn on axis "B"
        text_anchor = ax["B"].get_window_extent()

        locale_text = {
            "EN": "Solid lines indicate that the source\n"
            + "is visible. The visibility window is\n"
            + f"limited to a range between {self.min_alt} deg\n"
            + f"and {self.max_alt} deg.",
            "DE": "Durchgezogene Linien zeigen, dass die Quelle\n"
            + "sichtbar ist. Das Sichtbarkeitsfenster ist\n"
            + f"zwischen {self.min_alt}° und {self.max_alt}°"
            + "limitiert.",
        }

        locale_text_large_array = {
            "EN": " For array layouts with\n"
            + "more than 10 stations, only the first\n"
            + "station (ID 0) is shown.",
            "DE": " Für Anordnungen mit mehr als\n"
            + "10 Stationen wird nur die erste Station\n"
            + "(Index 0) gezeigt.",
        }

        if self.descr_text:
            text = locale_text[self.plot_locale]

            if self.location.size > 10:
                text += locale_text_large_array[self.plot_locale]

            ax["B"].annotate(
                text,
                (0.025, 0.025),
                xycoords=text_anchor,
                fontsize=12,
                va="bottom",
                ha="left",
                color="#41424C",
            )

        ax["B"].annotate(
            f"RA:\t{self.ra:.5f}\nDec:\t{self.dec:.5f}".expandtabs(),
            (0.025, 0.4),
            xycoords=text_anchor,
            fontsize=12,
            va="bottom",
            ha="left",
            fontfamily="monospace",
            color="#41424C",
        )
        if self.target_name:
            ax["B"].annotate(
                self.target_name,
                (0.025, 0.5),
                xycoords=text_anchor,
                fontsize=13,
                va="bottom",
                ha="left",
                fontfamily="monospace",
                color="#232023",
            )

    def plot(
        self,
        figsize: tuple[int, int] = (10, 5),
        colors: list = COLORS,
        *,
        descr_text: bool = True,
        legend: bool = True,
        locale: Literal["EN", "DE"] = "EN",
    ) -> tuple:
        """Plots the visibility of the source at the given
        time range. Also plots the positions of the sun and moon
        if set to ``True``.

        Parameters
        ----------
        figsize : tuple[int, int], optional
            Figure size. Width, height in inches.
            Default: (10, 5)
        colors : list, optional
            List of colors. If nothing provided a default
            list of colors is used. Default: None
        descr_text : bool, optional
            Whether to put a descriptive text next to the plot.
            Default: True
        legend : bool, optional
            If True, show legend in the plot. Default: True

        Returns
        -------
        tuple
            Figure and axis objects.
        """
        self.descr_text = descr_text
        self.legend = legend
        self.plot_locale = locale

        colors = iter(colors)

        fig, ax = plt.subplot_mosaic(
            "AB",
            layout="constrained",
            figsize=figsize,
            width_ratios=[3, 2],
        )

        for i, source_pos in enumerate(self.source_pos.values()):
            color = next(colors)

            mask = np.logical_and(
                source_pos.alt > self.min_alt * u.deg,
                self.max_alt * u.deg > source_pos.alt,
            )
            visible = source_pos.alt.copy()
            visible.value[~mask] = np.nan

            ax["A"].plot(
                self.dates,
                source_pos.alt,
                linestyle=":",
                color=color,
            )
            ax["A"].plot(
                self.dates,
                visible,
                lw=4,
                color=color,
                label=f"{i}",
            )

        ax["A"].plot()

        self._plot_config(ax)

        return fig, ax

    def _time_delta(self, r1, r2):
        latest_start = max(r1.start, r2.start)
        earliest_end = min(r1.end, r2.end)

        delta = earliest_end - latest_start

        return max(0, delta)

    def get_optimal_date(self, print_result: bool = False) -> list:
        """Computes the best date to observe the target source.
        Returns a list of three :class:`~pandas.Timestamp` where the
        first and last are the best date :math:`\pm` `obs_length / 2`.

        Parameters
        ----------
        print_result : bool, optional
            If `True`, also prints the result. Default: `False`

        Returns
        -------
        result : list
            List of :class:`~pandas.Timestamp`.
        """
        times = dict()
        t_range = namedtuple("t_range", ["start", "end"])

        for key, val in self.source_pos.items():
            maximum = np.max(val.alt)
            if maximum > self.max_alt * u.deg or maximum < self.min_alt * u.deg:
                continue
            idx_max = np.argmax(val.alt)
            delta = datetime.timedelta(hours=self.obs_length / 2)
            times[key] = [
                self.dates[idx_max] - delta,
                self.dates[idx_max],
                self.dates[idx_max] + delta,
            ]

        # Reindex times in case a key was skipped due to alt check.
        # This would be the case if, e.g. key 0 and 3  did not pass
        # the check resulting in times = {1: ..., 2: ..., 4: ...}.
        # This reindexing then creates times = {0: ..., 1: ..., 2: ...}.
        times = dict(enumerate(times.values()))

        dt = np.zeros([len(times), len(times)])
        for i, key_i in enumerate(times):
            for j, key_j in enumerate(times):
                r1 = t_range(
                    start=times[key_i][0].to_datetime64(),
                    end=times[key_i][-1].to_datetime64(),
                )
                r2 = t_range(
                    start=times[key_j][0].to_datetime64(),
                    end=times[key_j][-1].to_datetime64(),
                )

                dt[i, j] = self._time_delta(r1, r2)

        if dt.sum(axis=0).size == 0:
            raise ValueError(
                "The source is not visible with the chosen parameters. "
                "Radiotools was not able to determine the optimal "
                "observation date!"
            )

        result = times[int(np.argmax(dt.sum(axis=0)))]

        if print_result:
            print("")
            tab = Table(title="*** Best observation time ***")
            tab.add_column("Station ID", justify="right", style="cyan")
            tab.add_column("Obs. time start")
            tab.add_column("Obs. time midpoint")
            tab.add_column("Obs. time end")

            tab.add_row(
                f"{np.argmax(dt.sum(axis=0))}",
                result[0].strftime("%Y-%m-%d %H:%M:%S"),
                result[1].strftime("%Y-%m-%d %H:%M:%S"),
                result[2].strftime("%Y-%m-%d %H:%M:%S"),
            )
            console = Console()
            console.print(tab)
            print("")

        return result



from contextlib import contextmanager

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


@contextmanager
def assert_num_figures():
    n_fig_prev = plt.gcf().number
    yield
    n_fig_after = plt.gcf().number
    assert n_fig_prev < n_fig_after


def test_source_visibility():
    from radiotools.visibility import SourceVisibility

    sv = SourceVisibility(
        target="M87",
        date="2024-10-03",
        location="vlba",
        obs_length=4.0,
        frame="icrs",
        print_optimal_date=False,
    )

    with assert_num_figures():
        sv.plot()

    dates = sv.get_optimal_date()

    expected_dates = [
        pd.Timestamp("2024-10-03 16:52:58"),
        pd.Timestamp("2024-10-03 18:52:58"),
        pd.Timestamp("2024-10-03 20:52:58"),
    ]

    assert dates == expected_dates


def test_source_visibility_alt_restrictions():
    from radiotools.visibility import SourceVisibility

    alts = np.array([np.arange(0, 95, 5), np.arange(90, -5, -5)])

    opt_dates = [
        pd.Timestamp("2024-10-03 16:52:58"),
        pd.Timestamp("2024-10-03 18:52:58"),
        pd.Timestamp("2024-10-03 20:52:58"),
    ]

    results = []

    for i in np.arange(alts.shape[1]):
        try:
            results.append(
                SourceVisibility(
                    target="M87",
                    date="2024-10-03",
                    location="vlba",
                    obs_length=4.0,
                    frame="icrs",
                    min_alt=alts[0][i],
                    max_alt=alts[1][i],
                ).get_optimal_date()
                == opt_dates
            )
        except ValueError:
            results.append(False)

    assert np.all(results[:5])
    assert np.all(np.logical_not(results[5:]))



import subprocess as sp
from pathlib import Path

import toml


class WSClean:
    """Wrapper class for WSClean.

    Parameters
    ----------
    ms : str or Path
        Path to the measurement set.
    clean_config : dict or str or Path
        Either a dict containing the wsclean options or
        a path to a toml file containing the options.
        See note.
    create_skymodel : bool, optional
        Whether to create a skymodel before cleaning the image.
        Default: False.
    save_config : bool or str or Path, optional
        If True, save the config to a file relative to the file
        this class is called from. If given a path, the config
        is saved to that location instead. Default: False.

    Notes
    -----
    A valid config may look like the following:

    >>> clean_config = {
    >>>     'multiscale': True,
    >>>     'mgain': 0.5,
    >>>     'file_name': 'path/to/measurement_set',
    >>>     'data_column': 'DATA',
    >>>     'gain': 0.03,
    >>>     'weight': 'briggs 0',
    >>>     'mf_weighting': False,
    >>>     'size': [1024, 1024],
    >>>     'scale': '0.1masec',
    >>>     'pol': 'I',
    >>>     'niter': 5000000,
    >>>     'auto_threshold': 0.5,
    >>>     'auto_mask': 3,
    >>>     'padding': 1.3,
    >>>     'mem': 30,
    >>>     'verbose': True,
    >>> }
    """

    def __init__(
        self,
        ms: str | Path,
        clean_config: dict | str | Path = None,
        create_skymodel: bool = False,
        save_config: bool | str | Path = False,
    ) -> None:
        """Wrapper class for WSClean.

        Parameters
        ----------
        ms : str or Path
            Path to the measurement set.
        clean_config : dict or str or Path
            Either a dict containing the wsclean options or
            a path to a toml file containing the options.
            See note.
        create_skymodel : bool, optional
            Whether to create a skymodel before cleaning the image.
            Default: False.
        save_config : bool or str or Path, optional
            If True, save the config to a file relative to the file
            this class is called from. If given a path, the config
            is saved to that location instead. Default: False.
        """
        if not isinstance(clean_config, dict | str | Path):
            raise ValueError(
                "Please provide EITHER a dict object or a path"
                " to an existing configuration toml file."
            )

        if not Path(ms).is_dir():
            raise OSError(f"Measurement set {Path(ms).absolute()} does not exist.")

        self.ms = ms

        if isinstance(clean_config, str | Path):
            if not Path(clean_config).is_file():
                raise OSError(f"File {Path(clean_config).absolute()} does not exist.")

            with open(clean_config) as toml_file:
                clean_config = toml.load(toml_file)

        if save_config:
            self._save_config(clean_config, save_config)

        _clean_config = clean_config.copy()

        _clean_config["name"] = _clean_config.pop("file_name")
        _clean_config["data-column"] = _clean_config.pop("data_column")
        _clean_config["auto-threshold"] = _clean_config.pop("auto_threshold")
        _clean_config["auto-mask"] = _clean_config.pop("auto_mask")

        # We need to handle the case if size is an int and not a list.
        # This assumes, however, that x- and y- size are the same
        if isinstance(_clean_config["size"], int):
            _clean_config["size"] = [_clean_config["size"]] * 2

        # mf-weighting and no-mf-weighting are flags without arguments.
        # Since the wrapper only takes the bool mf_weighting as input,
        # we have to handle both cases and pass the respective WSClean
        # flags with an empty string as "argument".
        if _clean_config["mf_weighting"] is False:
            del _clean_config["mf_weighting"]
            _clean_config["no-mf-weighting"] = ""
        else:
            del _clean_config["mf_weighting"]
            _clean_config["mf-weighting"] = ""

        # Same as with mf-weighting, WSClean has two flags -- verbose
        # and quiet -- to handle output during cleaning. The wrapper
        # only takes the bool verbose as input, as such we have to
        # handle both cases.
        if _clean_config["verbose"] is False:
            del _clean_config["verbose"]
            _clean_config["quiet"] = ""
        else:
            del _clean_config["verbose"]

        # multiscale is another case of a flag without any arguments.
        # As such, we have to either delete the key and value pair,
        # or replace the bool with an empty string.
        if _clean_config["multiscale"] is False:
            del _clean_config["multiscale"]
        else:
            _clean_config["multiscale"] = ""

        self._clean_config = _clean_config

        if create_skymodel:
            self.skymodel_kwargs = _clean_config.copy()
            self.skymodel_kwargs["name"] += f"_{self.skymodel_kwargs['pol']}_skymodel"
            self.skymodel_kwargs["data-column"] = "DATA"
            self.skymodel_kwargs["auto-threshold"] = 1
            self.skymodel_kwargs["auto-mask"] = 3
            self.skymodel_kwargs["niter"] = 1000000

            self.create_skymodel()

    def create_skymodel(self) -> None:
        """Creates the skymodel using WSClean if ``create_skymodel``
        is set to ``True`` when initializing the class.
        """
        wsclean_opts = "wsclean "
        for key, val in self.skymodel_kwargs.items():
            if isinstance(val, list):
                wsclean_opts += f"-{key} {val[0]} {val[1]} "
            else:
                wsclean_opts += f"-{key} {val} "

        wsclean_opts += str(self.ms)
        sp.run(wsclean_opts, shell=True)

        print(f"Saved to {self.skymodel_kwargs['name']}<...>.fits")

    def clean_image(self) -> None:
        """Cleans the image using WSClean."""
        self._clean_config["name"] += "_" + self._clean_config["pol"]

        wsclean_opts = "wsclean "
        for key, val in self._clean_config.items():
            if isinstance(val, list):
                wsclean_opts += f"-{key} {val[0]} {val[1]} "
            else:
                wsclean_opts += f"-{key} {val} "

        wsclean_opts += self.ms
        sp.run(wsclean_opts, shell=True)

        print(f"Saved to {self._clean_config['name']}<...>.fits")

    def _save_config(self, _clean_config: dict, output_file: bool | str | Path) -> None:
        """Saves the config if ``save_config`` is set to ``True``
        when initializing the class.
        """
        if not isinstance(output_file, str | Path):
            output_file = Path(_clean_config["file_name"]).name
            output_file += f"_{_clean_config['pol']}" + "_config.toml"

        with open(output_file, "w") as toml_file:
            toml.dump(_clean_config, toml_file)



import numpy as np

_PREFIXES = dict.fromkeys(["microarcsecond", "µas"], (3.6e9, "µas"))
_PREFIXES.update(
    dict.fromkeys(["milliarcsecond", "marcsec", "masec", "mas"], (3.6e6, "mas"))
)
_PREFIXES.update(dict.fromkeys(["arcsecond", "arcsec", "asec", "as"], (3.6e3, "as")))


def px2radec(
    header: dict,
    xlim: tuple | None = None,
    ylim: tuple | None = None,
    num_ticks: int | tuple = 9,
    unit: str = "mas",
    ax=None,
):
    """Converts pixels to RA/Dec relative to source reference pixel.
    Expects the pixel increment to be in units of deg.

    Parameters
    ----------
    header : dict or astropy.io.fits.header.Header
        Header corresponding to the data.
    xlim : tuple or None, optional
        Set the x-axis view limits. If set to ``None``, the full
        axis size is used. Default: ``None``
    ylim : tuple or None, optional
        Set the y-axis view limits. If set to ``None``, the full
        axis size is used. Default: ``None``
    num_ticks : int or tuple, optional
        Number of ticks on the axes. If given a tuple,
        element 0 corresponds to the number of ticks on the
        x-axis, while element 1 corresponds to the y-axis.
    unit : str, optional
        The unit of the ticks. Can be one of ``'microarcsecond'``,
        ``'µas'``, ``'milliarcsecond'``, ``'marcsec'``, ``'masec'``, ``'mas'``,
        ``'arcsecond'``, ``'arcsec'``, ``'asec'``, ``'as'``. Default: ``'mas'``
    ax : matplotlib.axis.Axis, optional
        Matplotlib axis object to apply the conversion to. This will call
        the ``ax.set()`` method. Default: ``None``
    """
    if unit not in _PREFIXES:
        raise ValueError(f"Unknown unit! Please provide one of {_PREFIXES.keys()}")

    px_incr = header["CDELT2"]
    ref_pos = header["CRPIX1"] - 1, header["CRPIX2"] - 1
    naxis = header["NAXIS1"], header["NAXIS2"]

    xlim = (ref_pos[0] - xlim[0], ref_pos[0] + xlim[1]) if xlim else (0, naxis[0])
    ylim = (ref_pos[1] - ylim[0], ref_pos[1] + ylim[1]) if ylim else (0, naxis[1])

    if not isinstance(num_ticks, tuple):
        num_ticks = num_ticks, num_ticks

    xticks = np.linspace(*xlim, num_ticks[0], dtype=int, endpoint=True)
    yticks = np.linspace(*ylim, num_ticks[1], dtype=int, endpoint=True)

    xticklabels = xticks - ref_pos[0]
    yticklabels = yticks - ref_pos[1]

    xshift = xticklabels[xticklabels >= 0][0]
    yshift = yticklabels[yticklabels >= 0][0]

    xticklabels -= xshift
    yticklabels -= yshift
    xticklabels = [
        f"{label * px_incr * _PREFIXES[unit][0]:.2f}" if label != 0 else "0.00"
        for label in xticklabels
    ]
    yticklabels = [
        f"{label * px_incr * _PREFIXES[unit][0]:.2f}" if label != 0 else "0.00"
        for label in yticklabels
    ]

    if ax:
        ax.set(
            xlim=xlim,
            ylim=ylim,
            xticks=xticks,
            yticks=yticks,
            xticklabels=xticklabels,
            yticklabels=yticklabels,
            xlabel=rf"Relative RA $/\; \mathrm{{{_PREFIXES[unit][1]}}}$",
            ylabel=rf"Relative Dec $/\; \mathrm{{{_PREFIXES[unit][1]}}}$",
        )
    return xlim, ylim, xticks, yticks, xticklabels, yticklabels



import torch
from astropy.io import fits
from numpy.typing import ArrayLike

from radiotools.utils import rms


def get_source_rms(image: ArrayLike, center: tuple[int, int], *, offset=75):
    """Get the root mean square (RMS) of a given source.

    Parameters
    ----------
    image : array_like, shape (N, M)
        Array containing image data of the source.
    center : tuple[int, int]
        Tuple of the center coordinates where the RMS
        is calculated for a region spanned by ``offset``.
    offset : int, optional
        Offset that spans a region around ``center``:

        .. code::
            x0 = center_x - offset
            x1 = center_x + offset

            y0 = center_y - offset
            y1 = center_y + offset

        Default: 75

    Returns
    -------
    rms : float
        RMS for the given source.
    """
    center_x = int(center[0])
    center_y = int(center[1])

    x0 = center_x - offset
    x1 = center_x + offset

    y0 = center_y - offset
    y1 = center_y + offset

    _rms = torch.from_numpy(rms(image[x0:x1, y0:y1]))

    return torch.mean(_rms)


def dynamic_range(path: str, *, offset: int = 75):
    """Computes the dynamic range of a source.

    Parameters
    ----------
    path : str or Path
        Path to a FITS file containing image data of the source.
    offset : int, optional
        Offset for the region around the center pixel of the source.
        Used to compute the RMS value of the source. Default: 75

    Returns
    -------
    dr : float
        Dynamic range of the source.
    """
    hdu = fits.open(path)

    image = hdu[0].data[0, 0, ...]
    center = hdu[0].header["CRPIX1"], hdu[0].header["CRPIX2"]

    _rms = get_source_rms(image, center, offset=75)

    dr = image.max() / _rms

    return dr



"""Tool to calculate and show source visibility for a given time,
location/array layout.
"""

import datetime

import click


def main():
    target = click.prompt(
        "Target [ICRS name, leave empty to enter RA/Dec]",
        type=str,
        default="",
        show_default=False,
    )
    from matplotlib.pyplot import show

    from radiotools.visibility.visibility import SourceVisibility

    if target == "" or target.isspace():
        ra = click.prompt("Right ascension / deg", type=float)
        dec = click.prompt("Declination / deg", type=float)
        target = (ra, dec)

    date = click.prompt(
        "Date [Format: YYYY-MM-DD HH:MM:SS; leave empty for current time]",
        type=str,
        default="",
        show_default=False,
    )

    obs_length = click.prompt(
        "Observation length [hours]", type=float, default=8.0, show_default=True
    )

    if date == "" or date.isspace():
        date = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    location = click.prompt("Location [array layout or an existing location]")

    vis = SourceVisibility(
        target=target,
        date=date,
        location=location,
        obs_length=obs_length,
        print_optimal_date=True,
    )

    plot = click.confirm("Plot visibility?", default=False)
    if plot:
        fig, _ = vis.plot()
        show(block=True)


if __name__ == "__main__":
    main()



import numpy as np
from numpy.testing import assert_array_equal, assert_raises


class TestPlottingUtils:
    def setup_class(self):
        self._HEADER = {
            "CDELT2": 2.7777778241006e-08,
            "CRPIX1": 1023.0,
            "CRPIX2": 1024.0,
            "NAXIS1": 2048,
            "NAXIS2": 2048,
        }

    def test_px2radec(self):
        from radiotools.plotting import px2radec

        _xlim = (0, 2048)
        _ylim = (0, 2048)
        _xticks = np.array([0, 256, 512, 768, 1024, 1280, 1536, 1792, 2048])
        _yticks = np.array([0, 256, 512, 768, 1024, 1280, 1536, 1792, 2048])
        _xticklabels = [
            "-102.40",
            "-76.80",
            "-51.20",
            "-25.60",
            "0.00",
            "25.60",
            "51.20",
            "76.80",
            "102.40",
        ]
        _yticklabels = [
            "-102.40",
            "-76.80",
            "-51.20",
            "-25.60",
            "0.00",
            "25.60",
            "51.20",
            "76.80",
            "102.40",
        ]

        (xlim, ylim, xticks, yticks, xticklabels, yticklabels) = px2radec(self._HEADER)

        assert xlim == _xlim
        assert ylim == _ylim
        assert_array_equal(xticks, _xticks)
        assert_array_equal(yticks, _yticks)
        assert_array_equal(xticklabels, _xticklabels)
        assert_array_equal(yticklabels, _yticklabels)

    def test_px2radec_xylim(self):
        from radiotools.plotting import px2radec

        _xlim = (922, 1122)
        _ylim = (923, 1123)
        _xticks = np.array([922, 947, 972, 997, 1022, 1047, 1072, 1097, 1122])
        _yticks = np.array([923, 948, 973, 998, 1023, 1048, 1073, 1098, 1123])
        _xticklabels = [
            "-10.00",
            "-7.50",
            "-5.00",
            "-2.50",
            "0.00",
            "2.50",
            "5.00",
            "7.50",
            "10.00",
        ]
        _yticklabels = [
            "-10.00",
            "-7.50",
            "-5.00",
            "-2.50",
            "0.00",
            "2.50",
            "5.00",
            "7.50",
            "10.00",
        ]

        (xlim, ylim, xticks, yticks, xticklabels, yticklabels) = px2radec(
            self._HEADER, xlim=(100, 100), ylim=(100, 100)
        )

        assert xlim == _xlim
        assert ylim == _ylim
        assert_array_equal(xticks, _xticks)
        assert_array_equal(yticks, _yticks)
        assert_array_equal(xticklabels, _xticklabels)
        assert_array_equal(yticklabels, _yticklabels)

    def test_px2radec_raise(self):
        from radiotools.plotting import px2radec

        assert_raises(ValueError, px2radec, self._HEADER, unit="kiloarcsecond")

    def test_px2radec_num_ticks(self):
        from radiotools.plotting import px2radec

        _xticklabels = [
            "-102.40",
            "-76.80",
            "-51.20",
            "-25.60",
            "0.00",
            "25.60",
            "51.20",
            "76.80",
            "102.40",
        ]
        _yticklabels = [
            "-102.40",
            "-76.80",
            "-51.20",
            "-25.60",
            "0.00",
            "25.60",
            "51.20",
            "76.80",
            "102.40",
        ]

        (_, _, _, _, xticklabels, yticklabels) = px2radec(
            self._HEADER, num_ticks=(9, 9)
        )

        assert_array_equal(xticklabels, _xticklabels)
        assert_array_equal(yticklabels, _yticklabels)

    def test_px2radec_axis(self):
        import matplotlib.pyplot as plt

        from radiotools.plotting import px2radec

        _xticklabels = [
            "-102.40",
            "-76.80",
            "-51.20",
            "-25.60",
            "0.00",
            "25.60",
            "51.20",
            "76.80",
            "102.40",
        ]
        _yticklabels = [
            "-102.40",
            "-76.80",
            "-51.20",
            "-25.60",
            "0.00",
            "25.60",
            "51.20",
            "76.80",
            "102.40",
        ]

        fig, ax = plt.subplots()

        px2radec(self._HEADER, ax=ax)

        assert_array_equal(
            [txt.get_text() for txt in ax.get_xticklabels()], _xticklabels
        )
        assert_array_equal(
            [txt.get_text() for txt in ax.get_yticklabels()], _yticklabels
        )



def test_layouts():
    from radiotools.layouts import Layout

    BASE_URL = "https://raw.githubusercontent.com/radionets-project/"
    BASE_URL += "pyvisgen/refs/heads/main/resources/layouts/"

    layout = Layout.from_url(BASE_URL + "eht.txt")

    assert len(layout.names) == 8

    names = ["ALMA50", "SMTO", "LMT", "Hawaii8", "PV", "PdBI", "SPT", "GLT"]
    assert set(layout.names) == set(names)

    assert layout.x.dtype == "float64"
    assert layout.y.dtype == "float64"
    assert layout.z.dtype == "float64"

    x = [2225037.1851, -1828796.2, -768713.9637]
    y = [-5441199.162, -5054406.8, -5988541.7982]
    z = [-2479303.4629, 3427865.2, 2063275.9472]
    dish_dia = [84.700000, 10.000000, 50.000000]
    el_low = [15.000000, 15.000000, 15.000000]
    el_high = [85.000000, 85.000000, 85.000000]
    sefd = [110.0, 11900.0, 560.0]
    altitude = [5030.0, 3185.0, 4640.0]

    assert set(layout.x[:3]) == set(x)
    assert set(layout.y[:3]) == set(y)
    assert set(layout.z[:3]) == set(z)
    assert set(layout.dish_dia[:3]) == set(dish_dia)
    assert set(layout.el_low[:3]) == set(el_low)
    assert set(layout.el_high[:3]) == set(el_high)
    assert set(layout.sefd[:3]) == set(sefd)
    assert set(layout.altitude[:3]) == set(altitude)



import shutil
import uuid
import warnings
from datetime import datetime
from pathlib import Path

from astropy.io import fits

try:
    from casatools import ms as MeasurementTool
except ModuleNotFoundError as e:
    raise ModuleNotFoundError(
        "Cannot import casatools. Please make sure "
        "you installed radiotools with the optional "
        "casa dependency (uv pip install 'radiosim[casa]')!"
    ) from e


class Measurement:
    """
    A tool to convert between FITS files and NRAO CASA measurement sets
    """

    def get_obs_time(self):
        """
        Returns the datetime at which the Observation was started.
        """

        return datetime.fromisoformat(self.get_fits()[0].header["DATE-OBS"])

    def save_as_ms(self, path, overwrite=False):
        """
        Saves the current measurement
        If the Measurement was not not created using a CASA measurement set,
        this will raise an exception.

        Parameters
        ----------
        path: str
        The path of the root of the measurement set

        overwrite: bool, optional
        Whether an existing measurement set should be overwritten.
        """

        root = Path(path)

        if root.exists():
            if overwrite:
                shutil.rmtree(root)
            else:
                warnings.warn(
                    f"The directory {root} already exists! If you want "
                    "to overwrite it set overwrite=True!",
                    stacklevel=2,
                )

        if not root.exists() or overwrite:
            ms = MeasurementTool()
            ms.fromfits(msfile=path, fitsfile=self._fits_path, nomodify=not overwrite)

    def save_as_fits(self, path, overwrite=False):
        """
        Saves the current measurement as a FITS file
        If the Measurement was not not created using a FITS file,
        this will raise an exception.

        Parameters
        ----------
        path: str
        The path of the FITS file

        overwrite: bool, optional
        Whether an existing FITS file should be overwritten.
        """

        file = Path(path)

        if file.is_file():
            if overwrite:
                file.unlink()
            else:
                warnings.warn(
                    f"The file {file} already exists! If you want "
                    "to overwrite it set overwrite=True!",
                    stacklevel=2,
                )

        if not file.is_file() or overwrite:
            self._ms.tofits(path, overwrite=overwrite)

    def get_ms(self):
        """Returns a copy of the current measurement
        as a NRAO CASA measurement set. There will be
        no permanent physical version of the measurement
        on the disk. If the Measurement was not created
        using a FITS file, this will try to return a
        measurement set saved in this Measurement. If no
        measurement set is present, this will raise an exception.
        """

        if not hasattr(self, "_fits_path"):
            return self._ms

        def gen_rng_dir():
            return Path(f"./temp_fits_{uuid.uuid4()}/")

        temp_path = gen_rng_dir()

        while temp_path.exists():
            temp_path = gen_rng_dir()

        self.save_as_ms(temp_path.absolute(), overwrite=False)

        ms = MeasurementTool()
        ms.open(temp_path)

        shutil.rmtree(temp_path)

        return ms

    def get_fits(self):
        """Returns a copy of the current measurement
        as a FITS file. If the Measurement was not not
        created using a CASA measurement set, this will try
        to return a FITS file saved in this Measurement.
        If no FITS file is present, this will raise an exception.
        """
        if not hasattr(self, "_ms_path"):
            return self._fits

        def gen_rng_file():
            return Path(f"./temp_fits_{uuid.uuid4()}.fits")

        temp_path = gen_rng_file()

        while temp_path.is_file():
            temp_path = gen_rng_file()

        self.save_as_fits(str(temp_path), overwrite=False)

        _fits = fits.open(temp_path)
        temp_path.unlink()

        return _fits

    @classmethod
    def from_fits(cls, fits_path):
        """Creates a Measurement from a FITS file

        Parameters
        ----------
        fits_path: str
            The path to the FITS file
        """

        cls = cls()
        cls._fits_path = fits_path

        cls._fits = fits.open(fits_path)

        return cls

    @classmethod
    def from_ms(cls, ms_path):
        """Creates a Measurement from a NRAO CASA
        measurement set

        Parameters
        ----------
        ms_path: str
            The path to the root of the measurement set
        """

        cls = cls()

        ms = MeasurementTool()
        ms.open(ms_path)
        cls._ms_path = ms_path
        cls._ms = ms

        return cls



import urllib
import uuid
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from astropy.coordinates import EarthLocation
from casacore.tables import table
from numpy.typing import ArrayLike

pd.options.display.float_format = "{:f}".format


class Layout:
    """
    A tool to convert radio telescope array layout config files between different types.

    """

    def get_baselines(self):
        """Returns an array containing the lengths of all
        unique (!) baselines in meters.
        """
        loc = np.array([self.x, self.y]).T
        baselines = np.array([])

        for i, vec1 in enumerate(loc):
            for vec2 in loc[i + 1 :]:
                baselines = np.append(
                    baselines,
                    np.linalg.norm(np.reshape(vec1 - vec2, (2, 1)), ord=2, axis=0),
                )

        return baselines

    def get_baseline_vecs(self):
        """Returns an array containing the vectors of all (!)
        baselines (including conjugates).
        """
        loc = np.array([self.x, self.y]).T
        baselines = np.array([])

        for vec1 in loc:
            for vec2 in loc:
                baselines = np.append(baselines, np.reshape(vec1 - vec2, (2, 1)))

        return np.reshape(baselines, (int(len(baselines) / 2), 2)).T

    def get_max_resolution(self, frequency):
        """
        Returns the maximal resolution of the layout
        at a given frequency in arcsec / px.

        Parameters
        ----------
        frequency : float or array_like
            The frequency at which the array is observing

        """
        return 3600 * 180 / np.pi * 3 * 1e8 / (frequency * np.max(self.get_baselines()))

    def get_station(self, name):
        """Returns all information about the station
        with the given name as a `pandas.Series`.

        Parameters
        ----------
        name : str
            The name of the station (antenna).
            This is case sensitive!

        Returns
        -------
        pd.Series
            Pandas series containing all information
            about the given station.
        """
        if name not in self.names:
            raise KeyError(
                "This station could not be found. Make sure "
                "you typed the name correctly (case sensitive)!"
            )

        df = self.get_dataframe()

        return df.loc[df["station_name"] == name]

    def get_dataframe(self):
        """Returns the layout data as a :class:`pandas.DataFrame`."""

        return pd.DataFrame(
            data={
                "x": self.x,
                "y": self.y,
                "z": self.z,
                "dish_dia": self.dish_dia,
                "station_name": self.names,
                "el_low": self.el_low,
                "el_high": self.el_high,
                "sefd": self.sefd,
                "altitude": self.altitude,
            }
        )

    def __str__(self):
        output = f"Configuration file loaded from: {self.cfg_path}"
        output += f"\nRelative to site: {self.rel_to_site}"
        output += f"\nNumber of antennas: {len(self.x)}"
        output += f"\nNumber of baselines: {np.sum([i for i in range(len(self.x))])}"
        output += f"\nLongest baseline: {np.max(self.get_baselines())} m"
        output += f"\nShortest baseline: {np.min(self.get_baselines())} m"
        output += f"\n\n{self.get_dataframe()}"

        return output

    def display(self):
        """
        Prints all information contained in the layout

        """
        print(self.__str__())

    def is_relative(self):
        return not (self.rel_to_site is None or self.rel_to_site == "")

    def as_relative(self, rel_to_site):
        """Returns a copy of the current layout in
        relative coordinates.

        Parameters
        ----------
        rel_to_site : str
            The name of the site the coordinates are
            supposed to be relative to. Has to be an
            existing site for
            :func:`astropy.coordinates.EarthLocation.of_site()`.
        """

        def gen_rng_file():
            return Path(f"./temp_cfg_{uuid.uuid4()}.cfg")

        temp_path = gen_rng_file()

        while temp_path.is_file():
            temp_path = gen_rng_file()

        self.save(temp_path, rel_to_site=rel_to_site)
        new_layout = Layout.from_pyvisgen(temp_path, rel_to_site=rel_to_site)

        temp_path.unlink()
        new_layout.cfg_path = self.cfg_path

        return new_layout

    def as_absolute(self):
        """Returns a copy of the current layout in absolute
        coordinates. Requires the layout to be in relative
        coordinates.
        """

        if not self.is_relative():
            raise TypeError(
                "This layout is not saved in relative coordinates and can therefore "
                "not be converted in absolute coordinates."
            )

        def gen_rng_file(increment=0):
            return Path(f"./temp_cfg_{uuid.uuid4()}.cfg")

        temp_path = gen_rng_file()

        i = 1
        while temp_path.is_file():
            temp_path = gen_rng_file(i)
            i += 1

        self.save(temp_path)
        new_layout = Layout.from_pyvisgen(temp_path)

        temp_path.unlink()
        new_layout.cfg_path = self.cfg_path

        return new_layout

    def plot_uv(
        self,
        save_to_file="",
        ref_frequency=None,
        plot_args=None,
        save_args=None,
        show_zeros=False,
    ):
        """Plots the uv-sampling (uv-plane) of the array.

        Parameters
        ----------
        save_to_file : str, optional
            The name of the file the plot should be saved to.
        plot_args : dict, optional
            Arguments to pass to the axis.scatter function
        save_args : dict, optional
            Arguments to pass to the figure.savefig function
        """
        if plot_args is None:
            plot_args = {"color": "royalblue", "alpha": 0.5}

        if save_args is None:
            save_args = {}

        baselines = self.get_baseline_vecs()

        if not show_zeros:
            nonzero_bl = np.linalg.norm(baselines, axis=0) != 0
            baselines = baselines[:, nonzero_bl]

        fig, ax = plt.subplots(1, 1, layout="constrained")

        if ref_frequency is not None:
            baselines /= 3e8 / ref_frequency

        ax.scatter(baselines[0], baselines[1], **plot_args)
        ax.set_xlabel("$u$ in m" if ref_frequency is None else "$u/\\lambda$")
        ax.set_ylabel("$v$ in m" if ref_frequency is None else "$v/\\lambda$")

        if save_to_file != "":
            fig.savefig(save_to_file, **save_args)

        return fig, ax

    def plot(
        self,
        save_to_file="",
        annotate=False,
        limits=None,
        plot_args=None,
        save_args=None,
    ):
        """Generates a plot of the arrangement of the layout.

        Parameters
        ----------
        save_to_file : str, optional
            The name of the file the plot should be saved to.
        annotate : bool, optional
            Whether to mark the stations with their respective names.
        limits : tuple of tuples, optional
            The x and y bounds (e.g. `((0,1), (0,1))`). Set tuple of one
            axis (x or y) to None to only limit the other axis.
        plot_args : dict, optional
            Arguments to pass to the axis.scatter function
        save_args : dict, optional
            Arguments to pass to the figure.savefig function
        """
        if plot_args is None:
            plot_args = {}

        if save_args is None:
            save_args = {}

        singular_alt = len(np.unique(self.altitude)) == 1

        options = {
            "color": "#f54254" if singular_alt else None,
            "cmap": "cividis" if not singular_alt else None,
            "c": self.altitude if not singular_alt else None,
        }

        fig, ax = plt.subplots(1, 1)

        im = ax.scatter(self.x, self.y, **options, **plot_args)

        if limits:
            if limits[0]:
                ax.set(xlim=limits[0])
            if limits[1]:
                ax.set(ylim=limits[1])

        if not singular_alt:
            fig.colorbar(im, ax=ax, label="Altitude")

        if annotate:
            for _, row in self.get_dataframe().iterrows():
                ax.annotate(
                    text=f"{row['station_name']}", xy=(row.x, row.y), fontsize=8
                )

        ax.set_xlabel(f"{'Relative' if self.is_relative() else 'Geocentric'} $x$ in m")
        ax.set_ylabel(f"{'Relative' if self.is_relative() else 'Geocentric'} $y$ in m")
        ax.set_box_aspect(1)

        if save_to_file != "":
            fig.savefig(save_to_file, **save_args)

        return fig, ax

    def save(self, path, fmt="pyvisgen", overwrite=False, rel_to_site=None):
        """
        Saves the layout to a layout file.

        Parameters
        ----------
        path : str
            The path of the file to save the array layout to.
        fmt : str, optional
            The layout format the output file is supposed to have
            (available: casa, pyvisgen) (default is pyvisgen).
        overwrite : bool, optional
            Whether to overwrite the file if it already exists
            (default is False).
        rel_to_site : str, optional
            The name of the site the coordinates are supposed to be saved relative to.
            Is ignored if `None` or empty or `fmt` is not set to 'pyvisgen'.
            Has to be an existing site for
            :func:`astropy.coordinates.EarthLocation.of_site()`.
        """

        FORMATS = ["casa", "pyvisgen"]

        file = Path(path)

        if file.exists():
            if overwrite:
                file.unlink()
            else:
                raise FileExistsError(
                    f"The file {file} already exists! If you want "
                    "to overwrite it set overwrite=True!"
                )

        data = []

        save_relative = not (rel_to_site is None or rel_to_site == "")

        nx, ny, nz = self.x, self.y, self.z

        if save_relative:
            # Is supposed to saved be in relative (local tangent plane) coordinates

            location = EarthLocation.of_site(rel_to_site)

            if self.is_relative():
                # ... and is already relative --> reconvert
                # to absolute (if not same site)

                prev_location = EarthLocation.of_site(self.rel_to_site)
                nx, ny, nz = itrf2loc(
                    *loc2itrf(
                        prev_location.x.value,
                        prev_location.y.value,
                        prev_location.z.value,
                        self.x,
                        self.y,
                        self.z,
                    ),
                    location.x.value,
                    location.y.value,
                    location.z.value,
                )
            else:
                nx, ny, nz = itrf2loc(
                    self.x,
                    self.y,
                    self.z,
                    location.x.value,
                    location.y.value,
                    location.z.value,
                )

        else:
            # Is supposed to be saved in absolute (geocentric) coordinates

            if self.is_relative():
                # ... but is relative --> convert to absolute
                prev_location = EarthLocation.of_site(self.rel_to_site)
                nx, ny, nz = loc2itrf(
                    prev_location.x.value,
                    prev_location.y.value,
                    prev_location.z.value,
                    self.x,
                    self.y,
                    self.z,
                )

        match fmt:
            case "pyvisgen":
                data.append(
                    "station_name X Y Z dish_dia el_low el_high SEFD altitude\n"
                )

                for i in range(0, len(self.x)):
                    row = map(
                        str,
                        [
                            self.names[i],
                            nx[i],
                            ny[i],
                            nz[i],
                            self.dish_dia[i],
                            self.el_low[i],
                            self.el_high[i],
                            self.sefd[i],
                            self.altitude[i],
                        ],
                    )
                    data.append(" ".join(row) + "\n")

            case "casa":
                if save_relative:
                    data.append(f"observatory={rel_to_site}")
                    data.append("coordsys=LOC (local tangent plane)")

                data.append("# X Y Z dish_dia station_name\n")

                for i in range(0, len(self.x)):
                    row = map(
                        str,
                        [
                            nx[i],
                            ny[i],
                            nz[i],
                            self.dish_dia[i],
                            self.names[i],
                        ],
                    )
                    data.append(" ".join(row) + "\n")

            case _:
                raise ValueError(
                    f"{fmt} is not a valid format! Possible "
                    f"formats are: {', '.join(FORMATS)}!"
                )

        with open(file, "w", encoding="utf-8") as f:
            f.writelines(data)

    @classmethod
    def from_casa(
        cls, cfg_path, el_low=15, el_high=85, sefd=0, altitude=0, rel_to_site=None
    ):
        """
        Import a layout from a NRAO CASA layout config.

        Parameters
        ----------
        cfg_path : str
            The path of the config file to import.
        el_low : float or array_like, optional
            The minimal elevation in degrees the telescope can be adjusted to.
            If provided as singular number all telescopes in the array will
            be assigned the same value.
        el_high : float or array_like, optional
            The maximal elevation in degrees the telescope can be adjusted to.
            If provided as singular number all telescopes in the array will
            be assigned the same value.
        sefd : float or array_like, optional
            The system equivalent flux density of the telescope.
            If provided as singular number all telescopes in the array will
            be assigned the same value.
        altitude : float or array_like, optional
            The altitude of the telescope.
            If provided as singular number all telescopes in the array will
            be assigned the same value.
        rel_to_site : str, optional
            The name of the site the coordinates are relative to.
            Is ignored if `None` or empty.
            Has to be an existing site for
            :func:`astropy.coordinates.EarthLocation.of_site()`.
        """

        df = pd.read_csv(
            cfg_path,
            delimiter="\s+",
            encoding="utf-8",
            skip_blank_lines=True,
            names=["x", "y", "z", "dish_dia", "station_name"],
            dtype={
                "x": float,
                "y": float,
                "z": float,
                "dish_dia": float,
                "station_name": str,
            },
            comment="#",
        )
        cls = cls()
        cls.cfg_path = cfg_path
        cls.rel_to_site = rel_to_site
        cls.x = df["x"]
        cls.y = df["y"]
        cls.z = df["z"]
        cls.dish_dia = df["dish_dia"]
        cls.names = df["station_name"]
        cls.el_low = np.repeat(el_low, len(cls.x)) if np.isscalar(el_low) else el_low
        cls.el_high = (
            np.repeat(el_high, len(cls.x)) if np.isscalar(el_high) else el_high
        )
        cls.sefd = np.repeat(sefd, len(cls.x)) if np.isscalar(sefd) else sefd
        cls.altitude = (
            np.repeat(altitude, len(cls.x)) if np.isscalar(altitude) else altitude
        )

        return cls

    @classmethod
    def from_pyvisgen(cls, cfg_path, rel_to_site=None):
        """Import a layout from a radionets pyvisgen layout config.

        Parameters
        ----------
        cfg_path : str
            The path of the config file to import.

        rel_to_site : str, optional
            The name of the site the coordinates are relative to.
            Is ignored if `None` or empty.
            Has to be an existing site for
            :func:`astropy.coordinates.EarthLocation.of_site()`.
        """

        df = pd.read_csv(
            cfg_path,
            delimiter="\s+",
            encoding="utf-8",
            skip_blank_lines=True,
            dtype={
                "station_name": str,
                "x": float,
                "y": float,
                "z": float,
                "dish_dia": float,
                "el_low": float,
                "el_high": float,
                "sefd": float,
                "altitude": float,
            },
        )
        df.columns = map(str.lower, df.columns)

        cls = cls()
        cls.names = df["station_name"]
        cls.cfg_path = cfg_path
        cls.rel_to_site = rel_to_site
        cls.x = df["x"]
        cls.y = df["y"]
        cls.z = df["z"]
        cls.dish_dia = df["dish_dia"]
        cls.el_low = df["el_low"]
        cls.el_high = df["el_high"]
        cls.sefd = df["sefd"]
        cls.altitude = df["altitude"]

        return cls

    @classmethod
    def from_measurement_set(
        cls,
        root_path: str,
        sefd: int | ArrayLike,
        altitude: int | ArrayLike,
        rel_to_site: str | None = None,
    ):
        antennas = table(root_path + "/ANTENNA/", ack=False)

        positions = antennas.getcol("POSITION")
        stations = antennas.getcol("STATION")
        dish_diameters = antennas.getcol("DISH_DIAMETER")

        df = pd.DataFrame(
            data={
                "station_name": stations,
                "x": positions[0],
                "y": positions[1],
                "z": positions[2],
                "dish_dia": dish_diameters,
                "el_low": 2,
                "el_high": 90,
                "sefd": np.ones_like(stations, dtype=np.uint64()) * sefd
                if np.isscalar(sefd)
                else np.asarray(sefd),
                "altitude": np.ones_like(stations, dtype=np.uint64()) * altitude
                if np.isscalar(altitude)
                else np.asarray(altitude),
            }
        )

        return Layout.from_dataframe(df=df, rel_to_site=rel_to_site)

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame, rel_to_site: str | None = None):
        """Import a layout from a given pandas DataFrame.

        Parameters
        ----------
        df : pandas.DataFrame
            DateFrame containing the layout.

        rel_to_site : str, optional
            The name of the site the coordinates are relative to.
            Is ignored is `None` or empty or `fmt`. Has to be an
            existing site for `astropy.coordinates.EarthLocation.of_site()`.
            Default: None
        """
        cls = cls()
        cls.names = df["station_name"]
        cls.cfg_path = "DataFrame"
        cls.rel_to_site = rel_to_site
        cls.x = df["x"]
        cls.y = df["y"]
        cls.z = df["z"]
        cls.dish_dia = df["dish_dia"]
        cls.el_low = df["el_low"]
        cls.el_high = df["el_high"]
        cls.sefd = df["sefd"]
        cls.altitude = df["altitude"]

        return cls

    @classmethod
    def from_url(cls, url: str, rel_to_site: str | None = None) -> "Layout":
        """Import a layout from a given URL.

        Parameters
        ----------
        url : str
            URL of the layout file.
        rel_to_site : str, optional
            The name of the site the coordinates are relative to.
            Is ignored if `None` or empty. Has to be an
            existing site for :func:`astropy.coordinates.EarthLocation.of_site()`
            Default: None
        """
        data = []
        for line in urllib.request.urlopen(url):
            data.append(line.decode("utf-8").split())

        df = pd.DataFrame(data)
        df = df.rename(columns=df.iloc[0]).drop(df.index[0])
        df.columns = map(str.lower, df.columns)
        df = df.astype(
            {
                "station_name": str,
                "x": float,
                "y": float,
                "z": float,
                "dish_dia": float,
                "el_low": float,
                "el_high": float,
                "sefd": float,
                "altitude": float,
            }
        )

        cls = cls()
        cls.names = df["station_name"]
        cls.cfg_path = url
        cls.rel_to_site = rel_to_site
        cls.x = df["x"]
        cls.y = df["y"]
        cls.z = df["z"]
        cls.dish_dia = df["dish_dia"]
        cls.el_low = df["el_low"]
        cls.el_high = df["el_high"]
        cls.sefd = df["sefd"]
        cls.altitude = df["altitude"]

        return cls


def loc2itrf(cx, cy, cz, locx=0.0, locy=0.0, locz=0.0):
    """Returns the given points locx, locy, locz, which are
    relative to a common central point cx, cy, cz as the
    absolute points on the earth.

    Modified version of a CASAtasks script
    https://open-bitbucket.nrao.edu/projects/CASA/repos/casa6/browse/casa5/gcwrap/python/scripts/simutil.py

    Parameters
    ----------
    locx: array_like or float
        The x-coordinate in relative coordinates
    locy: array_like or float
        The y-coordinate in relative coordinates
    locz: array_like or float
        The z-coordinate in relative coordinates
    cx: float
        The center's x-coordinate in WGS84 coordinates
    cy: float
        The center's y-coordinate in WGS84 coordinates
    cz: float
        The center's z-coordinate in WGS84 coordinates
    """

    lon, lat, alt = geocentric2geodetic(cx, cy, cz)

    locx, locy, locz = map(np.array, (locx, locy, locz))
    # from Rob Reid;  need to generalize to use any datum...
    phi, lmbda = map(np.deg2rad, (lat, lon))
    sphi = np.sin(phi)
    a = 6378137.0  # WGS84 equatorial semimajor axis
    b = 6356752.3142  # WGS84 polar semimajor axis
    ae = np.arccos(b / a)
    N = a / np.sqrt(1.0 - (np.sin(ae) * sphi) ** 2)

    factor = (N + locz + alt) * np.cos(phi) - locy * sphi

    clmb = np.cos(lmbda)
    slmb = np.sin(lmbda)

    cx = factor * clmb - locx * slmb
    cy = factor * slmb + locx * clmb
    cz = (N * (b / a) ** 2 + locz + alt) * sphi + locy * np.cos(phi)

    return cx, cy, cz


def itrf2loc(x, y, z, cx, cy, cz):
    """Returns the relative position of given points
    x, y, z to a common central point cx, cy, cz on the earth.

    Modified version of a CASAtasks script
    https://open-bitbucket.nrao.edu/projects/CASA/repos/casa6/browse/casa5/gcwrap/python/scripts/simutil.py

    Parameters
    ----------
    x: array_like or float
        The x-coordinate in WGS84 coordinates
    y: array_like or float
        The y-coordinate in WGS84 coordinates
    z: array_like or float
        The z-coordinate in WGS84 coordinates
    cx: float
        The center's x-coordinate in WGS84 coordinates
    cy: float
        The center's y-coordinate in WGS84 coordinates
    cz: float
        The center's z-coordinate in WGS84 coordinates
    """

    clon, clat, h = geocentric2geodetic(cx, cy, cz)

    ccoslon = np.cos(clon)
    csinlon = np.sin(clon)
    csinlat = np.sin(clat)
    ccoslat = np.cos(clat)

    if isinstance(x, float):  # weak
        x = [x]
        y = [y]
        z = [z]
    n = x.__len__()
    lat = np.zeros(n)
    lon = np.zeros(n)
    el = np.zeros(n)

    # do like MsPlotConvert
    for i in range(n):
        # translate w/o rotating:
        xtrans = x[i] - cx
        ytrans = y[i] - cy
        ztrans = z[i] - cz
        # rotate
        lat[i] = (-csinlon * xtrans) + (ccoslon * ytrans)
        lon[i] = (
            (-csinlat * ccoslon * xtrans)
            - (csinlat * csinlon * ytrans)
            + ccoslat * ztrans
        )
        el[i] = (
            (ccoslat * ccoslon * xtrans)
            + (ccoslat * csinlon * ytrans)
            + csinlat * ztrans
        )

    return lat, lon, el


def geocentric2geodetic(x, y, z):
    """Returns given WGS84 coordinates (x,y,z) as geodetic
    coordinates (longitude [deg], latitude [deg], altitude [meter])

    Parameters
    ----------
    x: array_like or float
        The x-coordinate in WGS84 coordinates
    y: array_like or float
        The y-coordinate in WGS84 coordinates
    z: array_like or float
        The z-coordinate in WGS84 coordinates
    """

    if np.isscalar(x):
        lon, lat, alt = EarthLocation.from_geocentric(x, y, z, "m").to_geodetic()
        lon = lon.deg
        lat = lat.deg
        alt = alt.value
    else:
        lon, lat, alt = np.array([]), np.array([]), np.array([])
        for i in range(0, len(x)):
            loc = EarthLocation.from_geocentric(x[i], y[i], z[i], "m")
            lon = np.append(lon, loc.lon.deg)
            lat = np.append(lat, loc.lon.deg)
            alt = np.append(alt, loc.height.value)

    return lon, lat, alt


def geodetic2geocentric(lon, lat, alt):
    """Returns given geodetic coordinates (longitude, latitude, altitude)
    coordinates as (x,y,z) in meters

    Parameters
    ----------
    lon: array_like or float
        The longitude in geodetic coordinates
    lat: array_like or float
        The latitude in geodetic coordinates
    alt: array_like or float
        The altitude in geodetic coordinates
    """

    if np.isscalar(lon):
        x, y, z = EarthLocation.from_geodetic(
            lon=lon, lat=lat, height=alt
        ).to_geocentric()
        x = x.value
        y = y.value
        z = z.value
    else:
        x, y, z = np.array([]), np.array([]), np.array([])
        for i in range(0, len(lon)):
            loc = EarthLocation.from_geodetic(lon=lon[i], lat=lat[i], height=alt[i])
            x = np.append(x, loc.x.value)
            y = np.append(y, loc.y.value)
            z = np.append(z, loc.z.value)
    return x, y, z



import numpy as np
from numpy.testing import assert_allclose

from radiotools.utils import rms


class TestRMS:
    def test_1d_array(self):
        """Test 1d array."""
        a = np.array([10, 0, 150, 40, 30, 20, -50, -60, 5])
        expected = np.sqrt(1 / 9 * 31625)

        assert_allclose(rms(a), expected)

    def test_2d_array_axis_none(self):
        """Test 2d array with axis=None."""
        a = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        expected = 5.627314338711377

        assert_allclose(rms(a, axis=None), expected)

    def test_2d_array_axis_0(self):
        """Test 2d array with axis=0."""
        a = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        expected = np.array(
            [
                4.69041575982343,
                5.5677643628300215,
                6.48074069840786,
            ]
        )

        assert_allclose(rms(a, axis=0), expected)

    def test_2d_array_axis_1(self):
        """Test 2d array with axis=1."""
        a = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        expected = np.array(
            [
                2.160246899469287,
                5.066228051190222,
                8.04155872120988,
            ]
        )

        assert_allclose(rms(a, axis=1), expected)

    def test_input_ndim_0(self):
        """Test input with ndim=0"""
        a = np.int16(10)

        expected = 10.0

        assert rms(a) == expected



"""Print information about radiotools and its cli tools."""

import importlib
import sys
from collections import OrderedDict
from importlib.metadata import distribution, requires
from textwrap import TextWrapper

import click
from packaging.requirements import Requirement

__all__ = ["main"]


@click.command()
@click.option(
    "--version",
    is_flag=True,
    show_default=True,
    default=False,
    help="Print version number",
)
@click.option(
    "--dependencies",
    is_flag=True,
    show_default=True,
    default=False,
    help="Print dependencies",
)
@click.option(
    "--tools",
    is_flag=True,
    default=False,
    help="Print available cli tools",
)
def main(version, dependencies, tools):
    if len(sys.argv) <= 1:
        with click.get_current_context() as ctx:
            click.echo(ctx.get_help())
        sys.exit(1)

    if version:
        _info_version()

    if dependencies:
        _info_dependencies()

    if tools:
        _info_tools()


def _info_version():
    """Print version info."""
    import radiotools

    print("\n*** radiotools version info ***\n")
    print(f"version: {radiotools.__version__}")
    print("")


def _info_dependencies():
    """Print info about dependencies."""
    print("\n*** radiotools core dependencies ***\n")

    req = requires("radiotools")

    deps = [dep for dep in req if "dev" not in dep]

    for dep in deps:
        req = Requirement(dep)
        print(f"{req.name:>20} -- {str(req.specifier)}")


def _info_tools():
    """Print info about command line tools"""
    # Adapted from ctapipe
    # https://github.com/cta-observatory/ctapipe/blob/main/src/ctapipe/tools/info.py
    print("\n*** radiotools cli-tools ***\n")
    print("The following can be executed by typing radiotools-<toolname>:")
    print("")
    tools = {
        ep.name: ep.value
        for ep in distribution("radiotools").entry_points
        if ep.group in "console_scripts"
    }

    description = OrderedDict()
    for name, val in tools.items():
        module_name, attr = val.split(":")
        module = importlib.import_module(module_name)
        if hasattr(module, "__doc__") and module.__doc__ is not None:
            try:
                descr = module.__doc__
                descr.replace("\n", "")
                description[name] = descr
            except Exception as e:
                description[name] = f"[Could not parse docstring: {e}]"
        else:
            description[name] = "[No Documentation. Please add a docstring]"

    wrapper = TextWrapper(width=80, subsequent_indent=" " * 35)

    for name, descr in sorted(description.items()):
        text = f"{name:<30s} -- {descr}"
        print(wrapper.fill(text))
        print("")
    print("")


if __name__ == "__main__":
    main()



import re

import numpy as np
import requests
from astropy.io import fits
from bs4 import BeautifulSoup
from numpy.typing import ArrayLike


def get_array_names(url: str) -> list[str]:
    """Fetches array names from a given URL
    leading to a directory.

    Parameters
    ----------
    url : str
        URL leading to a directory containing layout
        txt files.

    Returns
    -------
    layouts : list[str]
        List of available layouts.
    """
    r = requests.get(url)
    soup = BeautifulSoup(r.text, features="html.parser")

    layouts = list(set(re.findall(r"\b\w+\.txt\b", str(soup.div))))

    return layouts


def rms(a: ArrayLike, *, axis: int | None = 0):
    """Return an array of the root-mean-square (RMS) value of
    the passed array.

    Parameters
    ----------
    a : array_like
        Array of which to find the RMS.
    axis : int or None

    Returns
    -------
    rms : np.ndarray
        Array of rms values.
    """
    if np.ndim(a) == 0:
        axis = None

    return np.sqrt(np.mean(a**2, axis=axis))


def img2jansky(image: ArrayLike, header: fits.Header):
    """Converts an image from Jy/beam to Jy/px.

    Parameters
    ----------
    image : array_like
        Input image that is to be converted.
    header : :class:`astropy.io.fits.header.Header`
        FITS file header belonging to the respective image.

    Returns
    -------
    array_like
        Converted image in units of Jy/px.
    """
    return (
        4
        * image
        * np.log(2)
        * np.power(header["CDELT1"], 2)
        / (np.pi * header["BMIN"] * header["BMAJ"])
    )
